
module_names = [
  '_dummy_thread',
  '_thread',
  'abc',
  'aifc',
  'argparse',
  'array',
  'ast',
  'asynchat',
  'asyncio',
  'asyncore',
  'atexit',
  'audioop',
  'base64',
  'bdb',
  'binascii',
  'binhex',
  'bisect',
  'builtins',
  'bz2',
  'calendar',
  'cgi',
  'cgitb',
  'chunk',
  'cmath',
  'cmd',
  'code',
  'codecs',
  'codeop',
  'collections.abc',
  'collections',
  'colorsys',
  'compileall',
  'concurrent.futures',
  'configparser',
  'contextlib',
  'contextvars',
  'copy',
  'copyreg',
  'crypt',
  'csv',
  'ctypes',
  'curses.ascii',
  'curses.panel',
  'curses.textpad',
  'curses',
  'dataclasses',
  'datetime',
  'dbm',
  'decimal',
  'difflib',
  'dis',
  'distutils',
  'doctest',
  'dummy_threading',
  'email',
  'ensurepip',
  'enum',
  'errno',
  'faulthandler',
  'fcntl',
  'filecmp',
  'fileinput',
  'fnmatch',
  'formatter',
  'fractions',
  'ftplib',
  'functools',
  'gc',
  'getopt',
  'getpass',
  'gettext',
  'glob',
  'grp',
  'gzip',
  'hashlib',
  'heapq',
  'hmac',
  'html.entities',
  'html.parser',
  'html',
  'http.client',
  'http.cookiejar',
  'http.cookies',
  'http.server',
  'http',
  'imaplib',
  'imghdr',
  'imp',
  'importlib',
  'inspect',
  'io',
  'ipaddress',
  'itertools',
  'json',
  'keyword',
  'linecache',
  'locale',
  'logging.config',
  'logging.handlers',
  'logging',
  'lzma',
  'macpath',
  'mailbox',
  'mailcap',
  'marshal',
  'math',
  'mimetypes',
  'mmap',
  'modulefinder',
  'msilib',
  'msvcrt',
  'multiprocessing',
  'netrc',
  'nis',
  'nntplib',
  'numbers',
  'operator',
  'optparse',
  'os.path',
  'os',
  'ossaudiodev',
  'parser',
  'pathlib',
  'pdb',
  'pickle',
  'pickletools',
  'pipes',
  'pkgutil',
  'platform',
  'plistlib',
  'poplib',
  'posix',
  'pprint',
  'pty',
  'pwd',
  'py_compile',
  'pyclbr',
  'pydoc',
  'queue',
  'quopri',
  'random',
  're',
  'readline',
  'reprlib',
  'resource',
  'rlcompleter',
  'runpy',
  'sched',
  'secrets',
  'select',
  'selectors',
  'shelve',
  'shlex',
  'shutil',
  'signal',
  'site',
  'site',
  'smtpd',
  'smtplib',
  'sndhdr',
  'socket',
  'socketserver',
  'spwd',
  'sqlite3',
  'ssl',
  'stat',
  'statistics',
  'string',
  'stringprep',
  'struct',
  'subprocess',
  'sunau',
  'symbol',
  'symtable',
  'sys',
  'sysconfig',
  'syslog',
  'tabnanny',
  'tarfile',
  'telnetlib',
  'tempfile',
  'termios',
  'test.support.script_helper',
  'test.support',
  'test',
  'textwrap',
  'threading',
  'time',
  'timeit',
  'tkinter.scrolledtext',
  'tkinter.tix',
  'tkinter.ttk',
  'tkinter',
  'token',
  'tokenize',
  'trace',
  'traceback',
  'tracemalloc',
  'tty',
  'turtle',
  'types',
  'typing',
  'unicodedata',
  'unittest.mock',
  'unittest.mock',
  'unittest',
  'urllib.error',
  'urllib.parse',
  'urllib.request',
  'urllib.response',
  'urllib.robotparser',
  'urllib',
  'uu',
  'uuid',
  'venv',
  'warnings',
  'wave',
  'weakref',
  'webbrowser',
  'winreg',
  'winsound',
  'wsgiref',
  'xdrlib',
  'xml.dom.minidom',
  'xml.dom.pulldom',
  'xml.dom',
  'xml.etree.ElementTree',
  'xml.parsers.expat',
  'xml.sax.handler',
  'xml.sax.saxutils',
  'xml.sax.xmlreader',
  'xml.sax',
  'xmlrpc.client',
  'xmlrpc.server',
  'xmlrpc',
  'zipapp',
  'zipfile',
  'zipimport',
  'zlib',
]
